"""
Copyright (C) 2022 - Mahesh Patil <mbpatil@ee.iitb.ac.in>
This file is part of GSEIM.

GSEIM is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import sys
import os
import numpy as np

class slv:
    def __init__(self, cct_filename):
        self.l_ovdict_all = []
        self.l_filename_all = []
        self.l_slv_type = []

        flag_begin_slv = False
        flag_begin_outblock = False

        if not os.path.exists(cct_filename):
            print('slv: file', cct_filename, 'not found. Halting...')
            sys.exit()
        f_in = open(cct_filename, "r")

        for line1 in f_in:
           line = line1.replace("\n","")
           flag_blank = len(line.split()) == 0
           if flag_blank:
               continue
           if flag_begin_slv:
               if line.split()[0] == 'end_solve':
                   flag_begin_slv = False
                   self.l_ovdict_all.append(l_ovdict_one)
                   self.l_filename_all.append(l_filename_one)
               else:
                   if flag_begin_outblock:
                       if line.split()[0] == 'end_output':
                           flag_begin_outblock = False
                           l_ovdict_one.append(d_ov)
                       if line.split()[0].split('=')[0] == 'filename':
                           l_filename_one.append(line.split()[0].split('=')[1])
                       if line.split()[0].split('=')[0] == 'variables:':
                           l1 = line.split()
                           n1 = 0
                           for i in range(1, len(l1)):
                               d_ov[l1[i]] = n1
                               n1 += 1
                   else:
                       if line.split()[0].split('=')[0] == 'solve_type':
                           self.l_slv_type.append(line.split()[0].split('=')[1])
                       if line.split()[0] == 'begin_output':
                           flag_begin_outblock = True
                           d_ov = {}
           else:
               if line.split()[0] == 'begin_solve':
                   flag_begin_slv = True
                   flag_begin_outblock = False
                   l_ovdict_one = []
                   l_filename_one = []

        f_in.close()

    def get_index(self, i_slv, i_outblock, ov_name):
#       Number of solve blocks in the circuit file: N (index 0,1,..,(N-1))
#       Solve block k has N_k output blocks (index 0,1,...,(N_k-1)

        if i_slv > len(self.l_ovdict_all)-1:
            print('get_index: i_slv:', i_slv, 'is out of range. Halting...')
            sys.exit()
        if i_outblock > len(self.l_ovdict_all[i_slv])-1:
            print('get_index: i_outblock:', i_outblock, 'is out of range. Halting...')
            sys.exit()
        if not ov_name in self.l_ovdict_all[i_slv][i_outblock].keys():
            print('get_index: ov_name:', ov_name, 'not found. Halting...')
            sys.exit()
        k = self.l_ovdict_all[i_slv][i_outblock][ov_name]
        if self.l_slv_type[i_slv] in ['ssw', 'trns', 'elex_rkf', 'elex_dp']:
            return (k+1)
        else:
            return k

    def get_n_col(self, i_slv, i_outblock):
        if i_slv > len(self.l_ovdict_all)-1:
            print('get_index: i_slv:', i_slv, 'is out of range. Halting...')
            sys.exit()
        if i_outblock > len(self.l_ovdict_all[i_slv])-1:
            print('get_index: i_outblock:', i_outblock, 'is out of range. Halting...')
            sys.exit()
        n_col = len(self.l_ovdict_all[i_slv][i_outblock])
        if self.l_slv_type[i_slv] in ['ssw', 'trns', 'elex_rkf', 'elex_dp']:
            return (n_col+1)
        else:
            return n_col

def replace_strings_1(filename_in, filename_out, l):
    f_in = open(filename_in, "r")
    f_out = open(filename_out, "w")
    for line in f_in:
        for t in l:
            line = line.replace(t[0], t[1])
        f_out.write(line)
    f_in.close()
    f_out.close()

def check_limits_1(x, x_min, x_max):
    if x > x_max:
        print('check_limits_1: x:', x, '>', x_max, 'Halting...')
        sys.exit()
    if x < x_min:
        print('check_limits_1: x:', x, '<', x_min, 'Halting...')
        sys.exit()

def avg_rms_1(t, x, T):

# - compute average and rms
# - t: time, x: quantity of interest (these are numpy arrays)
# - T: period
# - return a list of 3 lists (t_avg, x_avg, x_rms) which can be directly
#   used for plotting
# - Note: avg and rms values are computed for each period

    eps = 1.0e-3*T
    n  = len(t)

    sum_avg = 0.0
    sum_rms = 0.0

    t_last = t[0]
    x_last = x[0]

    t0 = t[0]
    t_end = t0 + T

    x_avg = []
    x_rms = []
    flag_eps = False
    for i in range(1, n):
        t1 = t[i]
        x1 = x[i]
        if abs(t1-t_end) < eps:
            if not flag_eps:
                sum_avg += 0.5*(x_last + x1)*(t1 - t_last)
                sum_rms += 0.5*(x_last*x_last + x1*x1)*(t1 - t_last)
                t0 = t_end
                t_end = t0 + T
                t_last = t1
                x_last = x1

                x_avg.append(sum_avg/T)
                x_rms.append(np.sqrt(sum_rms/T))
                sum_avg = 0.0
                sum_rms = 0.0
                flag_eps = True
        elif t1 > t_end:
            x1a = x_last + ((x1-x_last)/(t1-t_last))*(t_end-t_last)
            sum_avg += 0.5*(x_last + x1a)*(t_end - t_last)
            sum_rms += 0.5*(x_last*x_last + x1a*x1a)*(t_end - t_last)
            t_last = t_end
            x_last = x1a
            t0 = t_end
            t_end = t0 + T

            x_avg.append(sum_avg/T)
            x_rms.append(np.sqrt(sum_rms/T))
            sum_avg = 0.0
            sum_rms = 0.0
            flag_eps = False
        else:
            sum_avg += 0.5*(x_last + x1)*(t1 - t_last)
            sum_rms += 0.5*(x_last*x_last + x1*x1)*(t1 - t_last)
            t_last = t1
            x_last = x1
            flag_eps = False

    n1 = len(x_avg)

    l_r1 = [0.0]*(n1+2);
    l_r2 = [0.0]*(n1+2);
    l_r3 = [0.0]*(n1+2);

    for i in range(n1):
        l_r1[i+1] = (T/2) + float(i)*T
        l_r2[i+1] = x_avg[i]
        l_r3[i+1] = x_rms[i]

    if n1 == 1:
        l_r1[0] = l_r1[1]
        l_r2[0] = l_r2[1]
        l_r3[0] = l_r3[1]

        l_r1[2] = l_r1[1]
        l_r2[2] = l_r2[1]
        l_r3[2] = l_r3[1]
    else:
        l_r1[0] = l_r1[1] - 0.5*(l_r1[2]-l_r1[1])
        l_r2[0] = l_r2[1] - 0.5*(l_r2[2]-l_r2[1])
        l_r3[0] = l_r3[1] - 0.5*(l_r3[2]-l_r3[1])

        l_r1[n1+1] = l_r1[n1] + 0.5*(l_r1[n1]-l_r1[n1-1])
        l_r2[n1+1] = l_r2[n1] + 0.5*(l_r2[n1]-l_r2[n1-1])
        l_r3[n1+1] = l_r3[n1] + 0.5*(l_r3[n1]-l_r3[n1-1])

    l = []
    l.append(l_r1)
    l.append(l_r2)
    l.append(l_r3)

    return l

def avg_rms_2(t, x, t_start, t_end, eps):

# - compute average and rms
# - t: time, x: quantity of interest (these are numpy arrays)
# - t_start to t_end: interval of interest (for the numbers to make sense,
#   the user should ensure that this interval includes one or more periods
#   in the steady state)
# - eps should be small (say, 1e-3 * T)
# - return a list of 3 lists (t_avg, x_avg, x_rms) which can be directly
#   used for plotting
# - Note: Although there is only one avg and and one rms value, we will
#   assign the same value to multiple time points within [t_start, t_end]

    sum_avg = 0.0
    sum_rms = 0.0

    t_last = t[0];
    y_last = x[0];

    n  = len(t)

    for i in range(n):
        t0 = t[i]
        y0 = x[i]

        if (t_end-t_last) > (-eps):
            if (t_last-t_start) > (-eps):
                if (t_end-t0) > (-eps):
                    delta_t = t0 - t_last
                    sum_avg += 0.5*delta_t*(y_last + y0)
                    sum_rms += 0.5*delta_t*(y_last*y_last + y0*y0)
                else:
                    y1 = y_last + ((y0-y_last)/(t0-t_last))*(t_end-t_last)
                    delta_t = t_end-t_last
                    sum_avg += 0.5*delta_t*(y_last + y1)
                    sum_rms += 0.5*delta_t*(y_last*y_last + y1*y1)
            else:
                if (t0-t_start) > (-eps):
                    y1 = y_last + ((y0-y_last)/(t0-t_last))*(t_start-t_last)
                    delta_t = t0-t_start
                    sum_avg += 0.5*delta_t*(y1 + y0)
                    sum_rms += 0.5*delta_t*(y1*y1 + y0*y0)
        t_last = t0
        y_last = y0

    x_avg = sum_avg/(t_end-t_start)
    x_rms = np.sqrt(sum_rms/(t_end-t_start))

    n1 = 10
    t_new = np.linspace(t_start, t_end, n1)
    l_r1 = []
    l_r2 = []
    l_r3 = []

    for i in range(n1):
      l_r1.append(t_new[i])
      l_r2.append(x_avg)
      l_r3.append(x_rms)

    l = []
    l.append(l_r1)
    l.append(l_r2)
    l.append(l_r3)
    return l

def min_max_1(t, x, t_start, t_end):

# - Find min/max values of x between t_start and t_end
# - return [x_min, x_max]

    x_min =  1.0e20
    x_max = -1.0e20

    n  = len(t)

    for i in range(1, n):
        t1 = t[i]
        x1 = x[i]
        if (t1 >= t_start) and (t1 <= t_end):
            if x1 < x_min:
                x_min = x1
            if x1 > x_max:
                x_max = x1

    l = []
    l.append(x_min)
    l.append(x_max)
    return l

def min_max(filename, col_t, col_x, t_start, t_end):

# - Find min/max values of x between t_start and t_end
# - col_t, col_x: column index for t and x
#   (column numbers are counted as 1,2,3,..., NOT 0,1,2,...)
# - return [x_min, x_max]

    data = np.loadtxt(filename)

    t = data[:, (col_t-1)]
    x = data[:, (col_x-1)]

    l = min_max_1(t, x, t_start, t_end)
    return l

def get_value_1(t, x, t_given):

# - t: time, x: quantity of interest (these are numpy arrays)
# - t_given: time at which we want to find x
# - return x_given

    n  = len(t)
    flag_found = False

    t_last = t[0]
    x_last = x[0]

    for i in range(1, n):
        t1 = t[i]
        x1 = x[i]

        if (t_last <= t_given) and (t1 >= t_given):
            flag_found = True
            x2 = x_last + (t_given-t_last)*(x1-x_last)/(t1-t_last)

        t_last = t1
        x_last = x1

    if not flag_found:
        print('get_value_1: t_given not found. Halting...')
        sys.exit()

    return x2

def fourier_coeff(filename, col_t, col_x, t_start, t_end, n_fourier):
    data = np.loadtxt(filename)

    t = data[:, (col_t-1)]
    x = data[:, (col_x-1)]

# - compute fourier coefficients and THD
# - col_t, col_x: column index for t and x
#   (column numbers are counted as 1,2,3,..., NOT 0,1,2,...)
# - t_start, t_end: time interval for x(t)
# - return
#   - list of coefficients
#   - THD

    coeff = [0.0]*(n_fourier+1)
    sum_fourier_real = [0.0]*(n_fourier+1)
    sum_fourier_imag = [0.0]*(n_fourier+1)
    sum_fourier_mag = [0.0]*(n_fourier+1)
    omg = [0.0]*(n_fourier+1)
    sum_rms_1 = 0.0

    T = t_end - t_start
    a2 = 0.5*T
    f_hz = 1.0/T
    for i in range(n_fourier+1):
        omg[i] = 2.0*float(i)*np.pi/T

    t_last = t[0];
    y_last = x[0];

    n = len(t)

    for i in range(n):
        t0 = t[i]
        y0 = x[i]

        if t_last <= t_end:
            if t_last >= t_start:
                if t0 <= t_end:
                    delta_t = t0 - t_last
                    for j in range(n_fourier+1):
                        wt1 = omg[j]*(t_last - t_start)
                        wt2 = omg[j]*(t0 - t_start)
                        z1 = y_last
                        z2 = y0
                        sum_fourier_real[j] += 0.5*delta_t*(np.cos(wt1)*z1 + np.cos(wt2)*z2)
                        sum_fourier_imag[j] += 0.5*delta_t*(np.sin(wt1)*z1 + np.sin(wt2)*z2)
                    sum_rms_1 += 0.5*delta_t*(y_last*y_last + y0*y0)
                else:
                    y1 = y_last + ((y0-y_last)/(t0-t_last))*(t_end-t_last)
                    delta_t = t_end-t_last
                    for j in range(n_fourier+1):
                        wt1 = omg[j]*(t_last - t_start)
                        wt2 = omg[j]*(t_end - t_start)
                        z1 = y_last
                        z2 = y1
                        sum_fourier_real[j] += 0.5*delta_t*(np.cos(wt1)*z1 + np.cos(wt2)*z2)
                        sum_fourier_imag[j] += 0.5*delta_t*(np.sin(wt1)*z1 + np.sin(wt2)*z2)
                    sum_rms_1 += 0.5*delta_t*(y_last*y_last + y1*y1)
            else:
                if t0 > t_start:
                    y1 = y_last + ((y0-y_last)/(t0-t_last))*(t_start-t_last)
                    delta_t = t0-t_start
                    for j in range(n_fourier+1):
                        wt1 = 0.0
                        wt2 = omg[j]*(t0 - t_start)
                        z1 = y1
                        z2 = y0
                        sum_fourier_real[j] += 0.5*delta_t*(np.cos(wt1)*z1 + np.cos(wt2)*z2)
                        sum_fourier_imag[j] += 0.5*delta_t*(np.sin(wt1)*z1 + np.sin(wt2)*z2)
                    sum_rms_1 += 0.5*delta_t*(y1*y1 + y0*y0)
        t_last = t0
        y_last = y0

#   thd_sum = 0.0

    for i in range(n_fourier+1):
        b1 = sum_fourier_real[i]
        b2 = sum_fourier_imag[i]
        sum_fourier_mag[i] = np.sqrt((b1*b1)+(b2*b2))/a2

#       if i >= 2:
#           thd_sum += sum_fourier_mag[i]*sum_fourier_mag[i]

#   thd_first = sum_fourier_mag[1]
#   This is approximate; it depends on n_fourier, not desirable.
#   thd = np.sqrt(thd_sum)/thd_first

    coeff[0] = 0.5*sum_fourier_mag[0]
    for i in range(1, n_fourier+1):
        coeff[i] = sum_fourier_mag[i]

    arg1 = 2.0*(sum_rms_1/T) - 2.0*coeff[0]**2 - coeff[1]**2
    if coeff[1] < 1.0e-20:
        print('*** warning *** fourier_coeff: THD had div by zero.')
        print('*** setting thd_1 to zero')
        thd_1 = 0.0
    else:
        if arg1 < 1.0e-20:
            print('*** warning *** fourier_coeff: THD had sqrt of zero.')
            print('*** setting thd_1 to zero')
            thd_1 = 0.0
        else:
            thd_1 = np.sqrt(arg1)/coeff[1]

    return coeff, thd_1

def fourier_coeff_1A(t, x, t_start, t_end, t_eps, n_fourier):
#   similar to fourier_coeff_1B, but in this function, we provide t and x as
#   arguments, rather than reading from a file

# - compute fourier coefficients and THD
# - t_start, t_end: time interval for x(t)
# - return
#   - list of coefficients
#   - THD1, THD2

    coeff = [0.0]*(n_fourier+1)
    sum_fourier_real = [0.0]*(n_fourier+1)
    sum_fourier_imag = [0.0]*(n_fourier+1)
    sum_fourier_mag = [0.0]*(n_fourier+1)
    omg = [0.0]*(n_fourier+1)

    T = t_end - t_start
    a2 = 0.5*T
    f_hz = 1.0/T
    for i in range(n_fourier+1):
        omg[i] = 2.0*float(i)*np.pi/T

    n = len(t)

    flag_found_t_start = False;
    flag_found_t_end   = False;
    l_t = []
    l_x = []
    t_last = 0.0
    x_last = 0.0

    for i in range(n):
        t0 = t[i]
        x0 = x[i]

        if not flag_found_t_start:
            if t0 > (t_start-t_eps):
                flag_found_t_start = True
                if abs(t0-t_start) < t_eps:
                    l_t.append(t0)
                    l_x.append(x0)
                else:
#                   get the first point by interpolation:
                    if i == 0:
                        print('fourier_coeff_1A: t_start not found. Halting...')
                        sys.exit()
                    else:
                        x1 = x_last + (t_start-t_last)*(x0-x_last)/(t0-t_last)
                        l_t.append(t_start)
                        l_x.append(x1)

#                   add the current point also:
                    l_t.append(t0)
                    l_x.append(x0)

        if not flag_found_t_end:
            if t0 > (t_end-t_eps):
                flag_found_t_end = True
                if abs(t0-t_end) < t_eps:
                    l_t.append(t0)
                    l_x.append(x0)
                else:
#                   get the last point by interpolation:
                    x1 = x_last + (t_end-t_last)*(x0-x_last)/(t0-t_last)
                    l_t.append(t_end)
                    l_x.append(x1)
                break

        if flag_found_t_start:
            l_t.append(t0)
            l_x.append(x0)

        t_last = t0
        x_last = x0

#   print('fourier_coeff_1A: t, x:')
#   for i in range(len(l_t)):
#       print("%11.4E"%l_t[i], "%11.4E"%l_x[i])

    if not flag_found_t_start:
        print('fourier_coeff_1A: t_start not found. Halting...')
        sys.exit()
    if not flag_found_t_end:
        print('fourier_coeff_1A: t_end not found. Halting...')
        sys.exit()

    np_t = np.array(l_t) - t_start
    np_x = np.array(l_x)

#   compute:
#   sum_fourier_real[j]
#   sum_fourier_imag[j]
#   sum_rms_1

    for i in range(n_fourier+1):
        w = omg[i]
        f1 = (np.cos(w*np_t))*np_x
        f2 = (np.sin(w*np_t))*np_x
        sum_fourier_real[i] = np.trapz(f1, np_t)
        sum_fourier_imag[i] = np.trapz(f2, np_t)

    f3 = np_x*np_x
    sum_rms_1 = np.trapz(f3, np_t)

    for i in range(n_fourier+1):
        b1 = sum_fourier_real[i]
        b2 = sum_fourier_imag[i]
        sum_fourier_mag[i] = np.sqrt((b1*b1)+(b2*b2))/a2

    coeff[0] = 0.5*sum_fourier_mag[0]
    for i in range(1, n_fourier+1):
        coeff[i] = sum_fourier_mag[i]

    arg1 = 2.0*(sum_rms_1/T) - 2.0*coeff[0]**2 - coeff[1]**2
    if coeff[1] < 1.0e-20:
        print('*** warning *** fourier_coeff_1A: THD had div by zero.')
        print('*** setting thd_1 to zero')
        thd_1 = 0.0
    else:
        if arg1 < 1.0e-20:
            print('*** warning *** fourier_coeff_1A: THD had sqrt of zero.')
            print('*** setting thd_1 to zero')
            thd_1 = 0.0
        else:
            thd_1 = np.sqrt(arg1)/coeff[1]

    arg1 = 2.0*(sum_rms_1/T) - coeff[1]**2
    if coeff[1] < 1.0e-20:
        print('*** warning *** fourier_coeff_1A: THD had div by zero.')
        print('*** setting thd_1 to zero')
        thd_2 = 0.0
    else:
        if arg1 < 1.0e-20:
            print('*** warning *** fourier_coeff_1A: THD had sqrt of zero.')
            print('*** setting thd_1 to zero')
            thd_2 = 0.0
        else:
            thd_2 = np.sqrt(arg1)/coeff[1]

    return coeff, thd_1, thd_2

def fourier_coeff_1C(t, x, t_start, t_end, t_eps, n_fourier):
#   similar to fourier_coeff_1, but in this function, we provide t, x
#   as arguments, rather than reading from file

# - compute fourier coefficients and THD
# - t_start, t_end: time interval for x(t)
# - return
#   - list of coefficients
#   - THD

    coeff = [0.0]*(n_fourier+1)
    sum_fourier_real = [0.0]*(n_fourier+1)
    sum_fourier_imag = [0.0]*(n_fourier+1)
    sum_fourier_mag = [0.0]*(n_fourier+1)
    omg = [0.0]*(n_fourier+1)

    T = t_end - t_start
    a2 = 0.5*T
    f_hz = 1.0/T
    for i in range(n_fourier+1):
        omg[i] = 2.0*float(i)*np.pi/T

    n = len(t)

    flag_found_t_start = False;
    flag_found_t_end   = False;
    l_t = []
    l_x = []
    t_last = 0.0
    x_last = 0.0

    for i in range(n):
        t0 = t[i]
        x0 = x[i]

        if not flag_found_t_start:
            if t0 > (t_start-t_eps):
                flag_found_t_start = True
                if abs(t0-t_start) < t_eps:
                    l_t.append(t0)
                    l_x.append(x0)
                else:
#                   get the first point by interpolation:
                    if i == 0:
                        print('fourier_coeff_1C: t_start not found. Halting...')
                        sys.exit()
                    else:
                        x1 = x_last + (t_start-t_last)*(x0-x_last)/(t0-t_last)
                        l_t.append(t_start)
                        l_x.append(x1)

#                   add the current point also:
                    l_t.append(t0)
                    l_x.append(x0)

        if not flag_found_t_end:
            if t0 > (t_end-t_eps):
                flag_found_t_end = True
                if abs(t0-t_end) < t_eps:
                    l_t.append(t0)
                    l_x.append(x0)
                else:
#                   get the last point by interpolation:
                    x1 = x_last + (t_end-t_last)*(x0-x_last)/(t0-t_last)
                    l_t.append(t_end)
                    l_x.append(x1)
                break

        if flag_found_t_start:
            l_t.append(t0)
            l_x.append(x0)

        t_last = t0
        x_last = x0

#   print('fourier_coeff_1C: t, x:')
#   for i in range(len(l_t)):
#       print("%11.4E"%l_t[i], "%11.4E"%l_x[i])

    if not flag_found_t_start:
        print('fourier_coeff_1C: t_start not found. Halting...')
        sys.exit()
    if not flag_found_t_end:
        print('fourier_coeff_1C: t_end not found. Halting...')
        sys.exit()

    np_t = np.array(l_t) - t_start
    np_x = np.array(l_x)

#   compute:
#   sum_fourier_real[j]
#   sum_fourier_imag[j]
#   sum_rms_1

    for i in range(n_fourier+1):
        w = omg[i]
        f1 = (np.cos(w*np_t))*np_x
        f2 = (np.sin(w*np_t))*np_x
        sum_fourier_real[i] = np.trapz(f1, np_t)
        sum_fourier_imag[i] = np.trapz(f2, np_t)

    f3 = np_x*np_x
    sum_rms_1 = np.trapz(f3, np_t)

    for i in range(n_fourier+1):
        b1 = sum_fourier_real[i]
        b2 = sum_fourier_imag[i]
        sum_fourier_mag[i] = np.sqrt((b1*b1)+(b2*b2))/a2

    coeff[0] = 0.5*sum_fourier_mag[0]
    for i in range(1, n_fourier+1):
        coeff[i] = sum_fourier_mag[i]

    arg1 = 2.0*(sum_rms_1/T) - 2.0*coeff[0]**2 - coeff[1]**2
#   arg1 = 2.0*(sum_rms_1/T) - coeff[1]**2
    if coeff[1] < 1.0e-20:
        print('*** warning *** fourier_coeff_1C: THD had div by zero.')
        print('*** setting thd_1 to zero')
        thd_1 = 0.0
    else:
        if arg1 < 1.0e-20:
            print('*** warning *** fourier_coeff_1C: THD had sqrt of zero.')
            print('*** setting thd_1 to zero')
            thd_1 = 0.0
        else:
            thd_1 = np.sqrt(arg1)/coeff[1]

    return coeff, thd_1

def fourier_coeff_2A(t, x, t_start, t_end, t_eps, n_fourier):
#   similar to fourier_coeff_2, but in this function, we pass t, x rather than
#   the column numbers

# - compute fourier coefficients and THD
# - t_start, t_end: time interval for x(t)
# - return
#   - list of coefficients
#   - THD

    coeff = [0.0]*(n_fourier+1)
    coeff_cos = [0.0]*(n_fourier+1)
    coeff_sin = [0.0]*(n_fourier+1)
    sum_fourier_real = [0.0]*(n_fourier+1)
    sum_fourier_imag = [0.0]*(n_fourier+1)
    sum_fourier_mag = [0.0]*(n_fourier+1)
    omg = [0.0]*(n_fourier+1)

    T = t_end - t_start
    a2 = 0.5*T
    f_hz = 1.0/T
    for i in range(n_fourier+1):
        omg[i] = 2.0*float(i)*np.pi/T

    n = len(t)

    flag_found_t_start = False;
    flag_found_t_end   = False;
    l_t = []
    l_x = []
    t_last = 0.0
    x_last = 0.0

    for i in range(n):
        t0 = t[i]
        x0 = x[i]

        if not flag_found_t_start:
            if t0 > (t_start-t_eps):
                flag_found_t_start = True
                if abs(t0-t_start) < t_eps:
                    l_t.append(t0)
                    l_x.append(x0)
                else:
#                   get the first point by interpolation:
                    if i == 0:
                        print('fourier_coeff_2A: t_start not found. Halting...')
                        sys.exit()
                    else:
                        x1 = x_last + (t_start-t_last)*(x0-x_last)/(t0-t_last)
                        l_t.append(t_start)
                        l_x.append(x1)

#                   add the current point also:
                    l_t.append(t0)
                    l_x.append(x0)

        if not flag_found_t_end:
            if t0 > (t_end-t_eps):
                flag_found_t_end = True
                if abs(t0-t_end) < t_eps:
                    l_t.append(t0)
                    l_x.append(x0)
                else:
#                   get the last point by interpolation:
                    x1 = x_last + (t_end-t_last)*(x0-x_last)/(t0-t_last)
                    l_t.append(t_end)
                    l_x.append(x1)
                break

        if flag_found_t_start:
            l_t.append(t0)
            l_x.append(x0)

        t_last = t0
        x_last = x0

#   print('fourier_coeff_2A: t, x:')
#   for i in range(len(l_t)):
#       print("%11.4E"%l_t[i], "%11.4E"%l_x[i])

    if not flag_found_t_start:
        print('fourier_coeff_2A: t_start not found. Halting...')
        sys.exit()
    if not flag_found_t_end:
        print('fourier_coeff_2A: t_end not found. Halting...')
        sys.exit()

    np_t = np.array(l_t) - t_start
    np_x = np.array(l_x)

#   compute:
#   sum_fourier_real[j]
#   sum_fourier_imag[j]
#   sum_rms_1

    for i in range(n_fourier+1):
        w = omg[i]
        f1 = (np.cos(w*np_t))*np_x
        f2 = (np.sin(w*np_t))*np_x
        sum_fourier_real[i] = np.trapz(f1, np_t)
        sum_fourier_imag[i] = np.trapz(f2, np_t)

    f3 = np_x*np_x
    sum_rms_1 = np.trapz(f3, np_t)

    for i in range(n_fourier+1):
        b1 = sum_fourier_real[i]
        b2 = sum_fourier_imag[i]
        coeff_cos[i] = b1/a2
        coeff_sin[i] = b2/a2
        sum_fourier_mag[i] = np.sqrt((b1*b1)+(b2*b2))/a2

    coeff[0] = 0.5*sum_fourier_mag[0]
    for i in range(1, n_fourier+1):
        coeff[i] = sum_fourier_mag[i]

    arg1 = 2.0*(sum_rms_1/T) - 2.0*coeff[0]**2 - coeff[1]**2
    if coeff[1] < 1.0e-20:
        print('*** warning *** fourier_coeff_2A: THD had div by zero.')
        print('*** setting thd_1 to zero')
        thd_1 = 0.0
    else:
        if arg1 < 1.0e-20:
            print('*** warning *** fourier_coeff_2A: THD had sqrt of zero.')
            print('*** setting thd_1 to zero')
            thd_1 = 0.0
        else:
            thd_1 = np.sqrt(arg1)/coeff[1]

    return coeff, thd_1, coeff_cos, coeff_sin

def construct_fourier_component(k_fourier, coef_a, coef_b, t_start, T, n_periods, ndiv_per_period):
#   coef_a, b are real, imag fourier coeffs, respectively.

    if k_fourier > len(coef_a):
        print('construct_fourier_component: k_fourier is too large.')
        print('k_fourier:', k_fourier)
        print('len(coef_a):', len(coef_a))
        sys.exit()

    t = np.linspace(t_start, (t_start + n_periods*T), (n_periods*ndiv_per_period + 1))
#   print('construct_fourier_component: len(t):', len(t))
    w = 2.0*float(k_fourier)*np.pi/T
    t1 = t - t_start
    y = (np.cos(w*t1))*coef_a[k_fourier] + (np.sin(w*t1))*coef_b[k_fourier]
    return t, y

def cross_over_points_1(t, x, t_start, t_end, x0):
#   similar to cross_over_points; here, t, x are passed rather than column numbers
#   return two lists of time points:
#   - l_positive: x crossing x0 with a positive slope
#   - l_negative: x crossing x0 with a negative slope

    l_positive = []
    l_negative = []

    n = len(t)
    t_old = t[0]
    x_old = x[0]

    for i in range(1, n):
        t_new = t[i]
        x_new = x[i]
#       print('t_old:', "%11.4E"%t_old, 't_new:', "%11.4E"%t_new)
        if (x_new == x0) and (x_old == x0): continue
        if x_new >= x0:
            if x_old <= x0:
                if (t_old >= t_start) and (t_new <= t_end):
                    t_cross = t_old + (x0-x_old)*(t_new-t_old)/(x_new-x_old)
                    l_positive.append(t_cross)
        if x_new <= x0:
            if x_old >= x0:
                if (t_old >= t_start) and (t_new <= t_end):
                    t_cross = t_old + (x0-x_old)*(t_new-t_old)/(x_new-x_old)
                    l_negative.append(t_cross)
        t_old = t_new
        x_old = x_new

    return l_positive, l_negative

def interp_linear_1(x, y, ndiv):
    delx = (x[-1] - x[0])/ndiv
    x1 = x[0] + delx/2
    x2 = x[-1] - delx/2
    xnew = np.linspace(x1,x2,ndiv)
    ynew = np.interp(xnew, x, y)
    return delx, ynew

def delta_plot_1(x, a):
    x_max = np.max(x)
    x_min = np.min(x)
    delx = x_max - x_min
    plot_min_x = x_min - a*delx
    plot_max_x = x_max + a*delx
    return plot_min_x, plot_max_x
